/*========================================================================
 * Copyright (c) 2009 Sun Microsystems Laboratories
 * Copyright (c) 2009 Tampere University of Technology, Finland 
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without restriction,
 * including without limitation the rights to use, copy, modify, merge,
 * publish, distribute, sublicense, and/or sell copies of the Software,
 * and to permit persons to whom the Software is furnished to do so,
 * subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *======================================================================*/

// TODO: Functions that use parameter $super does not work!
// constructors that are global functions (for example klass()) does not work
// either.

/*======================================================================
 * QtClassBrowser.js: Simple JavaScript class browser
 *====================================================================*/

if (!Global) {
    var Global = this;
}

/*======================================================================
 * QtClassBrowser class
 *====================================================================*/

// Define the QtClassBrowser class as a subclass of QWidget
function QtClassBrowser(parent) {
    QWidget.call(this, parent);

    // Default window size
    this.defaultWidth = 600;
    this.defaultHeight = 300;

    // Set window (QWidget) attributes
    this.setWindowTitle("Lively Class Browser");
    this.resize(this.defaultWidth, this.defaultHeight);

    // Initialize browser state variables
    this.currentClass = null;
    this.currentMethod = null;

    // Create the widgets (three panes and one button) inside the browser
    this.classPane = new QTreeWidget();
    this.classPane.setHeaderLabels(["Classes"]);

    this.classMethodPane = new QTreeWidget();
    this.classMethodPane.setHeaderLabels(["Class Methods"]);

    this.instanceMethodPane = new QTreeWidget();
    this.instanceMethodPane.setHeaderLabels(["Instance Methods"]);

    this.saveChangesButton = new QPushButton(tr("Save changes"), this);
    this.saveChangesButton.enabled = false;

    this.codePane = new QTextEdit();

    // Create the horizontal splitter and add the panes to it
    var hSplitter = new QSplitter(this);
    hSplitter.addWidget(this.classPane);
    hSplitter.addWidget(this.classMethodPane);
    hSplitter.addWidget(this.instanceMethodPane);

    // Create the vertical splitter and add the panes to it
    var vSplitter = new QSplitter(Qt.Vertical, this);
    vSplitter.addWidget(hSplitter);
    vSplitter.addWidget(this.codePane);

    // Create the horizontal layout for code modification buttons.
    var buttonsLayout = new QHBoxLayout();
    buttonsLayout.addStretch();
    buttonsLayout.addWidget(this.saveChangesButton, 0, Qt.AlignRight);

    // Create layout that includes splitter layouts and buttons.
    var layout = new QVBoxLayout(this);
    layout.addWidget(vSplitter, 0, 0);
    layout.addLayout(buttonsLayout);
    layout.setContentsMargins(0, 0, 0, 0);
    this.setLayout(layout);

    // Initialize the state of the browser
    this.listClasses();
    this.initializeEventHandlers();

    return this;
}

// QtClassBrowser is a subclass of QWidget
QtClassBrowser.prototype = new QWidget();

/*======================================================================
 * Reflection interface
 *====================================================================*/

// Find all Lively JavaScript classes in the system
QtClassBrowser.prototype.listClasses = function() {
    var counter = 0;
    this.classPane.clear();

    for (var name in Global) {
        if (Class.isClass(Global[name])) {
            counter++;
            var treeItem = new QTreeWidgetItem(this.classPane);
            treeItem.setText(0, name);
        }
    }
    return counter;
}

// Find all the class methods of class 'thisClass'
QtClassBrowser.prototype.listClassMethods = function(thisClass) {
    var counter = 0;
    this.classMethodPane.clear();

    // Walk through class methods
    for (var name in thisClass) {

        if (Object.isFunction(thisClass[name])) {
            // Include only locally defined methods -- not those
            // inherited from superclasses.
            // TODO: We should traverse all the superclasses
            // in order to get a 100% reliable list of methods!
            if (thisClass[name] !== thisClass.superclass[name] &&
                name != "superclass") {
                counter++;
                var treeItem = new QTreeWidgetItem(this.classMethodPane);
                treeItem.setText(0, name);
            }
        }

    }

    return counter;
}

// Find all the instance methods of class 'thisClass'
QtClassBrowser.prototype.listInstanceMethods = function(thisClass) {
    var counter = 0;
    this.instanceMethodPane.clear();

    // Walk through instance methods
    for (var name in thisClass.prototype) {

        if (Object.isFunction(thisClass.prototype[name])) {
            // Include only locally defined methods -- not those
            // inherited from superclasses.
            // TODO: We should traverse all the superclasses
            // in order to get a 100% reliable list of methods!
            if (thisClass.prototype[name] !== thisClass.superclass.prototype[name] &&
                name != "constructor") {
                counter++;
                var treeItem = new QTreeWidgetItem(this.instanceMethodPane);
                treeItem.setText(0, name);
            }
        }

    }

    return counter;
}

/*==============================================================================
 * Event handlers
 *============================================================================*/

// Event handler for clicking on class names
QtClassBrowser.prototype.classChosen = function(treeItem) {
    var className = treeItem.text(0);
    this.currentClass = Global[className];
    if (this.currentClass) {
        this.listClassMethods(this.currentClass);
        this.listInstanceMethods(this.currentClass);
    } else {
        this.classMethodPane.clear();
        this.instanceMethodPane.clear();
    }
    this.codePane.plainText = " ";
    this.saveChangesButton.enabled = false;
}

// Event handler for clicking on class method names
QtClassBrowser.prototype.classMethodChosen = function(treeItem) {
    var methodName = treeItem.text(0);
    var thisClass = this.currentClass;
    this.currentMethodName = methodName;
    this.currentMethod = thisClass ? thisClass[methodName] : null;
    if (this.currentMethod) this.codePane.plainText = this.currentMethod;
    else this.codePane.plainText = "(no code)";

    this.saveChangesButton.enabled = true;
    this.methodIsClassMethod = true;

    // Remove selection(s) from the instance method pane
    this.instanceMethodPane.setCurrentItem(null);
}

// Event handler for clicking on instance method names
QtClassBrowser.prototype.instanceMethodChosen = function(treeItem) {
    var methodName = treeItem.text(0);
    var thisClass = this.currentClass;
    this.currentMethodName = methodName;
    this.currentMethod = thisClass ? thisClass.prototype[methodName] : null;
    if (this.currentMethod) this.codePane.plainText = this.currentMethod;
    else this.codePane.plainText = "(no code)";

    this.saveChangesButton.enabled = true;
    this.methodIsClassMethod = false;

    // Remove selection(s) from the class method pane
    this.classMethodPane.setCurrentItem(null);
}

// Event handler for saving changes of code
QtClassBrowser.prototype.saveChangesPressed = function() {
    // Create/update method
    var newMethod;
    try {
        newMethod = eval(this.codePane.plainText);
    }
    catch (e) {
        if (e instanceof SyntaxError) {
            var errorDialog = QMessageBox.warning(this, tr(""), tr(e.message));
        } else {
            var errorDialog = QMessageBox.warning(this, tr(""), tr(e));
        }
        return;
    }
    // Set/add method to class
    if (!newMethod) {
        print("ERROR: Unable to create new method from the following code:\n" + this.codePane.plainText);
        return;
    } else {
        if (!this.currentMethodName ||
            !this.currentClass) {
            print("No method name or class given!");
            return;
        }
        if (this.methodIsClassMethod) {
            this.currentClass[this.currentMethodName] = newMethod;
            // Reload code
            this.codePane.plainText = this.currentClass[this.currentMethodName];
        } else {
            this.currentClass.prototype[this.currentMethodName] = newMethod;
            // Reload code
            this.codePane.plainText = this.currentClass.prototype[this.currentMethodName];
        }
    }
}

// Event handler initialization
QtClassBrowser.prototype.initializeEventHandlers = function() {

    // Enable clicking on class names
    this.classPane.itemClicked.connect(this, this.classChosen);

    // Enable clicking on method names
    this.classMethodPane.itemClicked.connect(this, this.classMethodChosen);
    this.instanceMethodPane.itemClicked.connect(this, this.instanceMethodChosen);

    // Enable clicking of "Save changes" button
    this.saveChangesButton["clicked"].connect(this, this.saveChangesPressed);

}

/*==============================================================================
 * Instantiation
 *============================================================================*/

// var QtClassBrowser = new QtClassBrowser();
// QtClassBrowser.show();
// QCoreApplication.exec();

