/*========================================================================
 * Copyright (c) 2009 Tampere University of Technology, Finland
 * 
 * Permission is hereby granted, free of charge, to any person 
 * obtaining a copy of this software and associated documentation 
 * files (the "Software"), to deal in the Software without restriction, 
 * including without limitation the rights to use, copy, modify, merge, 
 * publish, distribute, sublicense, and/or sell copies of the Software, 
 * and to permit persons to whom the Software is furnished to do so, 
 * subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.  
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY 
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *======================================================================*/
 
// QtEventCal is a mashup application presenting information about 
// incoming music events. The concert and gig data is fetched from Last.fm 
// web service and the data is presented on a Google map. 
//
// TODO:
// * Add an artist-based search, i.e. one can "follow" an artist on his/her tour.
// * Add a calendar-based filtering (for example show gigs for a certain date 
//   or until a date)
// * Fix column widths etc.


function tr(s) { return s; }


function QtEventCal(parent) {
	QWidget.call(this, parent);

	// Add a text label with a brief presentation of the application.
	this.aboutApp = new QLabel(this);
	this.aboutApp.text = 'Type in the name of a location to search for music events:';

	// Add a field for the search text and a push button
	this.searchLineEdit = new QLineEdit(this);
	this.searchButton = new QPushButton("Search", this);
	// When the search button is clicked or the return button pressed, the 
	// events are fetched (based on the name of the location). 
	this.searchButton["clicked()"].connect(this, this.getEvents);
	this.searchLineEdit["returnPressed()"].connect(this, this.getEvents);
	// Add a layout for the search text field and push button
	this.searchLayout = new QHBoxLayout();
	this.searchLayout.addWidget(this.searchLineEdit, 0, 0);
	this.searchLayout.addWidget(this.searchButton, 0, 0);

	// QWebView handles showing a Google map
	this.view = new QWebView();
	this.view.load(new QUrl('http://www.cs.tut.fi/~nyrhinen/index.html'));
	this.view["loadFinished(bool)"].connect(this, this.loadFinished);

	// QTreeWidget contains a list of upcoming concerts and gigs (if found) 
	// located near the location used in the search. 
	this.eventTreeWidget = new QTreeWidget(this);
	var headerLabels = [];
	headerLabels.push(tr("Event"));
	headerLabels.push(tr("Artist(s)"));
	headerLabels.push(tr("Venue"));
	headerLabels.push(tr("Street"));
	headerLabels.push(tr("City"));
	headerLabels.push(tr("Date"));

	// Set the created labels as column headers
    this.eventTreeWidget.setHeaderLabels(headerLabels);
    // Make the header columns resizable by the user. 
    this.eventTreeWidget.header().setResizeMode(QHeaderView.Interactive);
    // Connect an action to a click of an item (the view is updated when an item
    // is clicked in the tree). 
	this.eventTreeWidget["itemClicked(QTreeWidgetItem*, int)"].connect(this, 
		this.itemClicked);

	// a vertical splitter for the webview and the splitter with the treewidget
	var vSplitter = new QSplitter(Qt.Vertical, this);
	vSplitter.addWidget(this.view);
	vSplitter.addWidget(this.eventTreeWidget);

	// The vertical main layout to which the widgets and other layouts are added
	this.actionLayout = new QVBoxLayout(this);
	this.actionLayout.addWidget(this.aboutApp, 0, 0);
	this.actionLayout.addLayout(this.searchLayout);
	this.actionLayout.addWidget(vSplitter, 0, 0);

	// Set the size, layout and title for the window	
	this.resize(640, 480);
    this.setLayout(this.actionLayout);
	this.setWindowTitle(tr("QtEventCal"));
}




// The main widget is of type QWidget
QtEventCal.prototype = new QWidget();




// Gets a list of upcoming events 
// The url is of format http://ws.audioscrobbler.com/2.0/?method=geo.getevents&location=madrid&api_key=b25b959554ed76058ac220b7b2e0a026
// Get your own API key at Last.fm!
// Demo API Key: dc369d73f68488889ef64ba06686f6dd
QtEventCal.prototype.getEvents = function() {
	// Empty the QTreeWidget before new items are added.
	this.eventTreeWidget.clear();
	
	// Form a URL with the typed in location as a search term. 
	var url = 'http://ws.audioscrobbler.com/2.0/?method=geo.getevents&location=' 
		+this.searchLineEdit.text + '&api_key=dc369d73f68488889ef64ba06686f6dd';
	// Obtain the events and proceed to handle the reply. 
	this.reqReply = this.downloadFile(url, this.replyFinished);
};




QtEventCal.prototype.getEventPage = function(page) {
	// Form a URL with the typed in location as a search term. 
	var url = 'http://ws.audioscrobbler.com/2.0/?method=geo.getevents&location=' 
		+this.searchLineEdit.text + '&api_key=dc369d73f68488889ef64ba06686f6dd'
		+'&page=' +page;
	// Obtain the events and proceed to handle the reply. 
	this.reqReply = this.downloadFile(url, this.replyFinished);
};





// Downloads the file pointed by the url
QtEventCal.prototype.downloadFile = function(url, callback) {
    // a QNetworkAccessManager handles sending requests and getting replies
    var accessManager = new QNetworkAccessManager();
    // An HTTP request is sent 
    var reqReply = accessManager.get(new QNetworkRequest(new QUrl(url)));
    // When the reply is obtained, the execution proceeds to the callback 
    // function.
    reqReply.finished.connect(this, callback);
    // the reply is returned
    return reqReply;
};




// Processes the contents of the reply
QtEventCal.prototype.replyFinished = function() {
    // store the read bytes
    var pageData = new QByteArray();
    // Read the data from the reply 
    pageData = this.reqReply.readAll();
    // New QXmlStreamReader for reading the feed is initialized with a 
    // QByteArray.
    var xml = new QXmlStreamReader(pageData);
    // Proceed to read the file contents. 
	this.read(xml);
};




// Handles reading the XML contents
QtEventCal.prototype.read = function(xml) {

	// Continue reading of the file until the end of the file.
    while (!xml.atEnd()) {
        xml.readNext();

        if (xml.isStartElement()) {
			if (xml.name() === 'events') {
				// Obtain the number of the current result page and the 
				// total number of result pages. 
				this.currPage = xml.attributes().value('page');
				this.totalPages = xml.attributes().value('totalpages');
			}
			if (xml.name() === 'event') {
				// Create a new QTreeWidgetItem for storing XML contents
				var item = new QTreeWidgetItem();
				var artists = []; // for storing of artists
			}
			if (xml.name() === 'title') {
				var title = xml.readElementText();
				item.setText(0, title);
			}
			if (xml.name() === 'artist') {
				var artist = xml.readElementText();
				artists.push(artist);
			}
			// headliner refers to the "main" artist of the event
			if (xml.name() === 'headliner') {
				//var headliner = xml.readElementText();
				//item.setText(1, headliner);
				
				var artistStr = ''; 
				// read all the artists contained within the array
				for (var i = 0; i < artists.length; i++) {
					artistStr +=  artists[i];
					if (i + 1 !== artists.length) {
					 	artistStr += ', ';
					}
				}
				// Set the item string containing the artists
				item.setText(1, artistStr);
				// clear the array
				artists.length = 0; 
			}
			// Venue name
			if (xml.name() === 'name') {
				var venue = xml.readElementText();
				item.setText(2, venue);
			}
			if (xml.name() === 'street') {
				var street = xml.readElementText();
				item.setText(3, street);
			}
			if (xml.name() === 'city') {
				var city = xml.readElementText();
				item.setText(4, city);
			}
			if (xml.name() === 'startDate') {
				var startDate = xml.readElementText();
				item.setText(5, startDate);
			}
			// If also a start time for an event is provided, add it to the 
			// QTreeWidget
			if (xml.name() === 'startTime') {
				var sDate = item.text(5); // Obtain the startDate
				var sTime = xml.readElementText();
				var startTime = sDate + ' at ' +sTime;
				item.setText(5, startTime);
			}
			if (xml.name() === 'lat') {
				var lat = xml.readElementText();
				item.setText(6, lat);
			}
			if (xml.name() === 'long') {
				var lng = xml.readElementText();
				item.setText(7, lng);
			}
			if (xml.name() === 'description') {
				var description = xml.readElementText();
				item.setText(8, description);
			}
			// Select only medium-sized images
			if (xml.name() === 'image' && 
				xml.attributes().value('size') === 'medium') {
				var imageUrl = xml.readElementText();
				item.setText(9, imageUrl);
			}
			if (xml.name() === 'website') {
				var website = xml.readElementText();
				item.setText(10, website);
			}
        }
        if (xml.isEndElement()) {
        	// If the contents of an event have been read (endtag is reached), 
        	// the item can be added to the tree.
           	if (xml.name() === 'event') {
	       		// Add the created item to the tree
				this.eventTreeWidget.addTopLevelItem(item);	
			}
			
			// If the end of the current page is reached and whether there are 
			// results left, proceed to download them.
			if (xml.name() === 'lfm') {
		    
				// If there are results left, proceed to download them
				if (parseInt(this.currPage, 10) < parseInt(this.totalPages, 10)) {
					var page = parseInt(this.currPage, 10) + 1;
					this.getEventPage(page);
				}
			}
			
        }
       
    }    
    return xml.error() == QXmlStreamReader.NoError;
};





// Sends a request to Google to geocode a location
QtEventCal.prototype.geoCode = function(location, callback) {
	var APIKey = 'ABQIAAAAjlFvU2tzDcjGvu_UgwNBmRQK3nvu5HLNMiWiNO-NKEHBcpToshSiHtBg-gl-shA2gvPoH2Y-BrW-qQ';
	var requestStr = "http://maps.google.com/maps/geo?q=" +
		location + "&output=csv&key=" + APIKey;
	var accessManager = new QNetworkAccessManager(this);
	accessManager["finished(QNetworkReply*)"].connect(this, callback);
	accessManager.get(new QNetworkRequest(new QUrl(requestStr)));
};




// Set map center point and zoom level
QtEventCal.prototype.setCenter = function(lat,lng,zoom) {
	var scriptString = 
	"map.setCenter(new GLatLng(" + lat + "," + lng + ")," + zoom + ");";
	this.view.page().mainFrame().evaluateJavaScript(scriptString);
};




QtEventCal.prototype.loadFinished = function(ok) {
	// If page loaded was the map page, do some initialization to the map
	if(this.view.page().mainFrame().url.toString() == 
			"http://www.cs.tut.fi/~nyrhinen/index.html") {
		var scriptString = 
			"var userClickedPoint = undefined;" +
			"var userClickedMarker = undefined;" +
			"var myClickListener = undefined;" +
			"var fromMarker = undefined;" +
			"var toMarker = undefined;" +
			"map.addControl(new GMapTypeControl());" +
			"map.addControl(new GLargeMapControl());" +
			"map.setCenter( new GLatLng(0,0),1 );" +
			"var markers = new Array();" +
			"myClickListener = GEvent.addListener(map, \"click\", function(marker, point) {" +
				"userClickedPoint = point;" +
				"userClickedMarker = marker;" +
			"});" +
			"GEvent.addListener(map, \"moveend\", function() {" +
		  		"userClickedPoint = undefined;" +
		  		"userClickedMarker = undefined;" +
			"});";
		this.view.page().mainFrame().evaluateJavaScript(scriptString);
	}
};




// When an item is clicked in the tree, search for the marker of the location 
// and set the centre of the map there.  
QtEventCal.prototype.itemClicked = function(item) {
	
	// Coordinates for the event and a zoom factor for the Google map
	var lat = item.text(6);
	var lng = item.text(7);
	var zoom = 13;

	// Set center to the point on the map 
	this.setCenter(lat,lng,zoom);

	// Form the description string for the marker with some 
	// some HTML formatting. The presented values are the title of the event, 
	// image of the artist, name(s) of the artist(s), venue and date.
	var descr = '<table><tr><td><h3>' +item.text(0) 
	+'</h3></td><td><img src=\'' 
	+item.text(9) +'\'></td></tr>'
	+'<tr><td>at ' +item.text(2) +'</td><td>&nbsp;</td></tr>'
	+'<tr><td>on ' +item.text(5) +'</td><td>&nbsp;</td></tr>'
	+'<tr><td>featured artists: ' +item.text(1) +'</td><td>&nbsp;</td></tr>'
	+'</table>';
	// Add a marker to the map
	this.addMarker(lat,lng,descr);				
};



// Adds a marker to a Google map
QtEventCal.prototype.addMarker = function(latitude, longitude, description) {
	var scriptString = 
	"var newMarker = new GMarker(new GLatLng(" + latitude + 
						"," + longitude + "),{ title: \""+description+"\" });" +
	"newMarker.bindInfoWindowHtml(newMarker.getTitle().toString());" +
	"markers.push(newMarker);" +
	"map.addOverlay(markers[markers.length-1]);";
	this.view.page().mainFrame().evaluateJavaScript(scriptString);
};





// Returns the contents of QByteArray as a string
QByteArray.prototype.toString = function() {
   ts = new QTextStream( this, QIODevice.ReadOnly );
   return ts.readAll();
};

 
/* To use the application in standalone mode, remove the comment around this code 
var wdgt = new QtEventCal();
wdgt.show();
QCoreApplication.exec();
*/

