/*========================================================================
 * Copyright (c) 2009 Tampere University of Technology, Finland
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without restriction,
 * including without limitation the rights to use, copy, modify, merge,
 * publish, distribute, sublicense, and/or sell copies of the Software,
 * and to permit persons to whom the Software is furnished to do so,
 * subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *======================================================================*/

PayazzoStopper.prototype.GRAVITY = 40.0;
PayazzoStopper.prototype.FALL_TIME = Math.sqrt(10 / PayazzoStopper.prototype.GRAVITY);

function PayazzoStopper(x, y, space, owner)
{
    this.x = x;
    this.y = y;
    this.space = space;
    this.owner = owner;
    this.falling = false;
}

PayazzoStopper.prototype.initCircle = function()
{
    this.circle = new PhysicsCircle(this.x, this.y + 10*this.space, 5, 0, 0, true);
    return this.circle;
}

PayazzoStopper.prototype.run = function(secm, coins, simulation)
{

    // Check if coins are falling
    if (this.falling) {
        this.falling_time += secm;
        while (this.falling_time > this.FALL_TIME) {
            this.falling_time -= this.FALL_TIME;
            this.falling_left --;
            this.owner.addCash(1);
            this.space ++;
            if (this.falling_left <= 0) {
                break;
            }
        }
        if (this.falling_left > 0) {
            this.circle.y = this.y + 10*this.space + this.falling_time*this.falling_time*this.GRAVITY;
            return;
        }
        this.falling = false;
        this.circle.y = this.y + 10*this.space;
    }

    // If there is no space, then do nothing
    if (this.space <= 0) {
        return;
    }

    var MAX_X_DISTANCE = 5.0;
    var MAX_Y_DISTANCE = 0.5;
    var MAX_Y_VELOCITY = 1.5;

    var next_free_y = this.y + (this.space-1) * 10;

    // Go coins through
    for (var coins_id = 0; coins_id < coins.length; coins_id ++) {
        var coin = coins[coins_id];
        // Skip coins that do not fill the landing criteria
        if (coin.x > this.x + MAX_X_DISTANCE ||
            coin.x < this.x - MAX_X_DISTANCE ||
            coin.y > next_free_y + MAX_Y_DISTANCE ||
            coin.y < next_free_y - MAX_Y_DISTANCE ||
            coin.yv < -MAX_Y_VELOCITY ||
            coin.yv > MAX_Y_VELOCITY) {
            continue;
        }
        coins[coins_id] = coins[coins.length-1];
        coins.length --;
        simulation.removeCircle(coin, true);
        this.space --;
        this.circle.y -= 10;
    }

}

PayazzoStopper.prototype.isActive = function()
{
    return this.falling;
}

PayazzoStopper.prototype.getDroppableNow = function(max_space)
{
    if (this.falling) {
        return 0;
    }
    return max_space - this.space;
}

PayazzoStopper.prototype.dropCoins = function(drops)
{
    if (!this.falling) {
        this.falling = true;
        this.falling_time = 0;
        this.falling_left = drops;
    } else {
        this.falling_left += drops;
    }
}

PayazzoStopper.prototype.dropIfFull = function(drops)
{
    if (this.space == 0 && !this.falling) {
        this.falling = true;
        this.falling_time = 0;
        this.falling_left = drops;
    }
}

PayazzoStopper.prototype.getPixelSpace = function()
{
    if (this.falling) {
        return this.space*10 + this.falling_time*this.falling_time*this.GRAVITY;
    }
    return this.space*10;
}

PayazzoPort = function(x, y, points)
{
    this.x = x;
    this.y = y;
    this.points = points;
}

PayazzoPort.prototype.checkForPasses = function(coins)
{
    var MAX_X_DISTANCE = 2.0;
    var MAX_Y_DISTANCE = 4.0;

    var passes = 0;
    for (var coins_id = 0; coins_id < coins.length; coins_id ++) {
        var coin = coins[coins_id];
        // Skip coins that have already passed through a port.
        if (coin.passed_a_port) {
            continue;
        }
        // Skip coins that are outside of port
        if (coin.x > this.x + MAX_X_DISTANCE ||
            coin.x < this.x - MAX_X_DISTANCE ||
            coin.y > this.y + MAX_Y_DISTANCE ||
            coin.y < this.y - MAX_Y_DISTANCE) {
            continue;
        }
        coin.passed_a_port = true;
        passes ++;
    }
    return this.points * passes;
}

QtPayazzo.prototype = new QWidget();

QtPayazzo.prototype.FPS = 24;
QtPayazzo.prototype.HISCORES_COUNT = 30;

function QtPayazzo(parent)
{

    QWidget.call(this, parent);

    this.pSim = new PhysicsSimulation(this, 320, 480);
    this.mainlayout = new QVBoxLayout();
    this.mainlayout.setContentsMargins(0, 0, 0, 0);
    this.layoutholder = null;

    // Load pixmaps
    this.files_loaded = false;
    var files_to_load = [ "http://lively.cs.tut.fi/images/payazzo/Payazzo.png",
                          "http://lively.cs.tut.fi/images/payazzo/PayazzoMask.png",
                          "http://lively.cs.tut.fi/images/payazzo/Coin.png",
                          "http://lively.cs.tut.fi/images/payazzo/5Coins.png",
                          "http://lively.cs.tut.fi/images/payazzo/16Coins.png",
                          "http://lively.cs.tut.fi/images/payazzo/NewGame.png",
                          "http://lively.cs.tut.fi/images/payazzo/InsertCoinHover.png",
                          "http://lively.cs.tut.fi/images/payazzo/InsertCoinPress.png",
                          "http://lively.cs.tut.fi/images/payazzo/CashOutHover.png",
                          "http://lively.cs.tut.fi/images/payazzo/CashOutPress.png" ];
    this.fileloader = new Fileloader();
    this.fileloader.loadFiles(files_to_load, this, this.fileLoaded);

    this.initBoard();

    this.setLayout(this.mainlayout);

    this.hiscores = [];
    while (this.hiscores.length < this.HISCORES_COUNT) {
        this.hiscores.push({ score: 0, name: "---"});
    }

    this.game_running = false;
    this.game_able_to_end = true;

    // Init timers
    this.timer = new QTimer(this);
    this.timer.timeout.connect(this, this.runGame);

}

QtPayazzo.prototype.getSizeHint = function()
{
    return new QSize(320, 480);
}

QtPayazzo.prototype.getMinimumSizeHint = function()
{
    return new QSize(320, 480);
}

QtPayazzo.prototype.keyPressEvent = function(event)
{
    if (!this.files_loaded) {
        return;
    }
    if (!this.game_running && !this.layoutholder) {
        this.startNewGame();
    }
}

QtPayazzo.prototype.mouseMoveEvent = function(event)
{
    if (!this.files_loaded) {
        return;
    }

    // If the game is not running, then do nothing.
    if (!this.game_running) {
        return;
    }

    // Check if mouse cursor hovers over buttons
    var need_update = false;
    if (event.x() >= 198 && event.x() < 262 &&
        event.y() >= 12 && event.y() < 36) {
        if (!(this.cashout & 1)) {
            this.cashout |= 1;
            need_update = true;
        }
    } else if (this.cashout & 1) {
        this.cashout ^= 1;
        need_update = true;
    }
    if (event.x() >= 284 && event.x() < 312 &&
        event.y() >= 12 && event.y() < 36) {
        if (!(this.insertcoin & 1)) {
            this.insertcoin |= 1;
            this.update();
        } else if (need_update) {
            this.update();
        }
    } else if (this.insertcoin & 1) {
        this.insertcoin ^= 1;
        this.update();
    } else if (need_update) {
        this.update();
    }

    // If there are no credits, then shooting is not possible
    if (this.credits <= 0) {
        return;
    }

    var ms = new Date().getTime();

    var SENSOR_TIME = 100;

    this.mousemoves.push({ x: event.x(), y: event.y(), time: ms });
    // Remove expired mousemoves
    var expiretime = ms - SENSOR_TIME;
    while (this.mousemoves.length > 2 && this.mousemoves[0].time < expiretime) {
        this.mousemoves.shift();
    }

    // If mouse was moved out of screen from left edge, then shoot coin
    if (event.x() < 0) {
        var oldest_mm = this.mousemoves[0];
        var x_move = event.x() - oldest_mm.x;
        var y_move = event.y() - oldest_mm.y;
        var ms_diff = ms - oldest_mm.time;
        // Coin is shooted if there was enough time difference and enough move
        // parts. Also time difference from last shooting time is checked
        if (ms_diff <= SENSOR_TIME && this.mousemoves.length >= 2 && this.last_shoot_time < ms - 500) {
            x_move /= ms_diff;
            y_move /= ms_diff;
            if (x_move > 160) x_move = 160;
            if (x_move < -160) x_move = -160;
            if (y_move > 160) y_move = 160;
            if (y_move < -160) y_move = -160;
            this.shootCoin(-5 + x_move * 45, -10 + y_move * 5);
            this.last_shoot_time = ms;
            this.credits --;
        }
    }
}

QtPayazzo.prototype.mousePressEvent = function(event)
{

    if (!this.files_loaded) {
        return;
    }

    if (!this.game_running) {
        if (!this.layoutholder) {
            this.startNewGame();
        }
        return;
    }

    // Check if one of buttons is pressed
    var need_update = false;
    if (event.x() >= 198 && event.x() < 262 &&
        event.y() >= 12 && event.y() < 36) {
        if (!(this.cashout & 2)) {
            this.cashout |= 2;
            need_update = true;
        }
    }
    if (event.x() >= 284 && event.x() < 312 &&
        event.y() >= 12 && event.y() < 36) {
        if (!(this.insertcoin & 2)) {
            this.insertcoin |= 2;
            this.update();
        } else if (need_update) {
            this.update();
        }
    } else if (need_update) {
        this.update();
    }
}

QtPayazzo.prototype.mouseReleaseEvent = function(event)
{
    if (!this.files_loaded) {
        return;
    }
    // If the game is not running, then do nothing.
    if (!this.game_running) {
        return;
    }
    // Check if mouse was over a pressed button when it was released
    var need_update = false;
    if (event.x() >= 198 && event.x() < 262 &&
        event.y() >= 12 && event.y() < 36) {
        if (this.cashout & 2) {
            this.cashout = 1;
            if (this.credits == 0 && this.cash > 0 && this.game_able_to_end) {
                this.showHiscore();
            }
            need_update = true;
        }
    }
    if (event.x() >= 284 && event.x() < 312 &&
        event.y() >= 12 && event.y() < 36) {
        if (this.insertcoin & 2) {
            this.insertcoin = 1;
            if (this.cash > 0) {
                this.cash --;
                this.credits ++;
            }
            this.update();
        } else if (need_update) {
            this.update();
        }
    } else if (need_update) {
        this.update();
    }

    this.cashout &= 1;
    this.insertcoin &= 1;
}

QtPayazzo.prototype.runGame = function(event)
{
    this.run(1.0 / this.FPS, false);
}

QtPayazzo.prototype.startNewGame = function()
{
    this.game_running = true;
    this.mouseTracking = true;
    this.mousemoves = [];
    this.last_shoot_time = 0;
    this.credits = 8;
    this.cashout = 0;
    this.insertcoin = 0;
    this.update();
}

QtPayazzo.prototype.run = function(secm)
{
    this.pSim.run(secm);

    // Go through stoppers and run them. This makes them fall or check if any
    // coins have landed on them.
    for (var stoppers_id = 0; stoppers_id < this.stoppers.length; stoppers_id ++) {
        var stopper = this.stoppers[stoppers_id];
        stopper.run(secm, this.coins, this.pSim);
    }

    // Check if special stoppers at sides are full.
    this.stoppers[this.stoppers.length-1].dropIfFull(5);
    this.stoppers[this.stoppers.length-2].dropIfFull(5);

    // Check if coins have falled out
    for (var coins_id = 0; coins_id < this.coins.length; coins_id ++) {
        var coin = this.coins[coins_id];
        if (coin.y > 480) {
            this.coins[coins_id] = this.coins[this.coins.length-1];
            this.coins.length --;
            this.pSim.removeCircle(coin, true);
        }
    }

    // Check if some coins have passed through ports
    for (var ports_id = 0; ports_id < this.ports.length; ports_id ++) {
        var port = this.ports[ports_id];
        var points = port.checkForPasses(this.coins);
        if (points > 0) {
            this.dropCoins(points);
        }
    }

    // Check if game can be marked as not active. This is a state where it is
    // possible to have game over and take cash out if there are no credits
    // left. Note, that test for active coins is done only if time from last
    // shooting is not very big. If enough time has passed from last coin
    // shooting, then it is not a good idea to check for active coins, since
    // some coins may be jammed.
    this.game_able_to_end = true;
    if (this.last_shoot_time + 1000 * 60 > (new Date).getTime()) {
        for (var coins_id = 0;
             coins_id < this.coins.length;
             coins_id ++) {
            var coin = this.coins[coins_id];
            if (coin.x < 25 || coin.x > 295 ||
                coin.y < 225 ||
                coin.xv < -75 || coin.xv > 75 ||
                coin.yv < -75 || coin.yv > 100) {
                this.game_able_to_end = false;
                break;
            }
        }
    }
    if (this.game_able_to_end) {
        for (var stoppers_id = 0;
             stoppers_id < this.stoppers.length;
             stoppers_id ++) {
            if (this.stoppers[stoppers_id].isActive()) {
                this.game_able_to_end = false;
                break;
            }
        }
    }
    if (this.game_able_to_end &&
        this.cash == 0 &&
        this.credits == 0) {
        this.game_running = false;
    }

    // If there are no coins, then check if coin stacks are not active. If
    // nothing active is found, then it means that timers can be stopped.
    if (this.coins.length == 0) {
        var actives_found = false;
        for (var stoppers_id = 0;
             stoppers_id < this.stoppers.length;
             stoppers_id ++) {
            if (this.stoppers[stoppers_id].isActive()) {
                actives_found = true;
                break;
            }
        }
        if (!actives_found) {
            this.timer.stop();
        }
    }

    this.update();

}

QtPayazzo.prototype.dropCoins = function(drops)
{
    var MAX_SIZE = 16;
    while (drops > 0) {
        // Pick the cell that has most of coins left.
        var most_coins_ids = [];
        var most_coins = 1;
        for (stoppers_id = 0;
             stoppers_id < this.stoppers.length - 2;
             stoppers_id ++) {
            var stopper = this.stoppers[stoppers_id];
            var able_to_drop = stopper.getDroppableNow(MAX_SIZE);
            if (able_to_drop > most_coins) {
                most_coins_ids = [];
                most_coins_ids.push(stoppers_id);
                most_coins = able_to_drop;
            } else if (able_to_drop == most_coins) {
                most_coins_ids.push(stoppers_id);
            }
        }
        if (most_coins_ids.length == 0) {
            break;
        }
        var most_coins_id = most_coins_ids[Math.floor(Math.random()*most_coins_ids.length)];
        var coins_to_drop = Math.min(drops, most_coins);
        this.stoppers[most_coins_id].dropCoins(coins_to_drop);
        drops -= coins_to_drop;
    }
    this.addCash(drops);
}

QtPayazzo.prototype.shootCoin = function(x_vel, y_vel)
{
    var new_coin = new PhysicsCircle(325, 80, 5, x_vel, y_vel, true);
    new_coin.render_x = new_coin.x;
    new_coin.render_y = new_coin.y;
    this.coins.push(new_coin);
    this.pSim.addCircle(new_coin, true);
    this.timer.start(1000.0 / this.FPS);
}

QtPayazzo.prototype.initBoard = function()
{

    this.coins = [];
    this.stoppers = [];
    this.ports = [];
    this.credits = 0;
    this.cash = 0;

    this.pSim.reset();

    // Ports
    for (var ports = 0; ports < 7; ports ++) {
        var x = (ports - 3) * 35 + 160;
        var y = 160;
        var points1 = [];
        points1.push(new QPointF(x+ -5,  y+ 8));
        points1.push(new QPointF(x+ -5,  y+ -8));
        points1.push(new QPointF(x+ -9, y+ -8));
        points1.push(new QPointF(x+ -11,  y+ 8));
        var points2 = [];
        points2.push(new QPointF(x+ 5,  y+ 8));
        points2.push(new QPointF(x+ 5,  y+ -8));
        points2.push(new QPointF(x+ 9, y+ -8));
        points2.push(new QPointF(x+ 11,  y+ 8));
        this.pSim.addPolygon(new PhysicsPolygon(points1));
        this.pSim.addPolygon(new PhysicsPolygon(points2));
    }
    this.ports.push(new PayazzoPort(55, 160, 2));
    this.ports.push(new PayazzoPort(90, 160, 3));
    this.ports.push(new PayazzoPort(125, 160, 8));
    this.ports.push(new PayazzoPort(160, 160, 20));
    this.ports.push(new PayazzoPort(195, 160, 8));
    this.ports.push(new PayazzoPort(230, 160, 3));
    this.ports.push(new PayazzoPort(265, 160, 2));

    // Edges
    var left_edge = [];
    left_edge.push(new QPointF(8, 0));
    left_edge.push(new QPointF(8, 640));
    this.pSim.addPath(new PhysicsPath(left_edge, true, true, false));
    var top_edge = [];
    top_edge.push(new QPointF(0, 48));
    top_edge.push(new QPointF(320, 48));
    this.pSim.addPath(new PhysicsPath(top_edge, true, true, false));
    var right_edge = [];
    right_edge.push(new QPointF(312, 0));
    right_edge.push(new QPointF(312, 70));
    right_edge.push(new QPointF(335, 70));
    right_edge.push(new QPointF(335, 90));
    right_edge.push(new QPointF(312, 90));
    right_edge.push(new QPointF(312, 640));
    this.pSim.addPath(new PhysicsPath(right_edge, true, true, false));

    // Edges of cells
    for (var cedges = 0; cedges < 13; cedges ++) {
        var x = (cedges - 6) * 14 + 160;
        var y = 240 + Math.abs(cedges - 6) * 10;
        var points = [];
        points.push(new QPointF(x+ -2, y+ 640));
        if (cedges == 0) {
            points.push(new QPointF(x+ -2,  y+ 59));
            points.push(new QPointF(x+ -14,  y+ 50));
            points.push(new QPointF(x+ -14,  y+ 10));
            points.push(new QPointF(x+ 0,  y+ 0));
            points.push(new QPointF(x+ 2, y+ 1));
        } else if (cedges == 12) {
            points.push(new QPointF(x+ -2,  y+ 1));
            points.push(new QPointF(x+ 0,  y+ 0));
            points.push(new QPointF(x+ 14,  y+ 10));
            points.push(new QPointF(x+ 14,  y+ 50));
            points.push(new QPointF(x+ 2,  y+ 59));
        } else {
            points.push(new QPointF(x+ -2,  y+ 1));
            points.push(new QPointF(x+ 0,  y+ 0));
            points.push(new QPointF(x+ 2, y+ 1));
        }
        points.push(new QPointF(x+ 2, y+ 640));
        this.pSim.addPath(new PhysicsPath(points, true, true, false));
    }

    // Special edges at both sides
    var sedges1 = [];
    sedges1.push(new QPointF(18, 640));
    sedges1.push(new QPointF(18, 220));
    sedges1.push(new QPointF(32, 210));
    sedges1.push(new QPointF(88, 250));
    sedges1.push(new QPointF(88, 260));
    sedges1.push(new QPointF(50, 287));
    sedges1.push(new QPointF(50, 360));
    sedges1.push(new QPointF(62, 369));
    sedges1.push(new QPointF(62, 640));
    this.pSim.addPath(new PhysicsPath(sedges1, true, true, false));
    var sedges2 = [];
    sedges2.push(new QPointF(320-18, 640));
    sedges2.push(new QPointF(320-18, 220));
    sedges2.push(new QPointF(320-32, 210));
    sedges2.push(new QPointF(320-88, 250));
    sedges2.push(new QPointF(320-88, 260));
    sedges2.push(new QPointF(320-50, 287));
    sedges2.push(new QPointF(320-50, 360));
    sedges2.push(new QPointF(320-62, 369));
    sedges2.push(new QPointF(320-62, 640));
    this.pSim.addPath(new PhysicsPath(sedges2, true, true, false));

    // Initialize coin stoppers

    // Cell stoppers
    for (var cstoppers = 0; cstoppers < 12; cstoppers ++) {
        var x = (cstoppers - 5.5) * 14 + 160;
        var y = 246 + Math.abs(cstoppers - 5.5) * 10;
        var new_stopper = new PayazzoStopper(x, y, 0, this);
        this.stoppers.push(new_stopper);
        this.pSim.addCircle(new_stopper.initCircle(), false);
    }

    // Special port stoppers
    var new_stopper1 = new PayazzoStopper(13, 320-10*5, 5, this);
    var new_stopper2 = new PayazzoStopper(320-13, 320-10*5, 5, this);
    this.pSim.addCircle(new_stopper1.initCircle(), false);
    this.pSim.addCircle(new_stopper2.initCircle(), false);
    this.stoppers.push(new_stopper1);
    this.stoppers.push(new_stopper2);

}

QtPayazzo.prototype.addCash = function(amount)
{
    this.cash += amount;
}

QtPayazzo.prototype.showHiscore = function()
{
    this.game_running = false;

    // Check if current score is good enough to be added to highscores
    var score_added = false;
    for (var hiscores_id = 0; hiscores_id < this.hiscores.length; hiscores_id ++) {
        if (this.hiscores[hiscores_id].score < this.cash) {
            this.hiscores.length ++;
            for (var hiscores_mover = this.hiscores.length-1;
                 hiscores_mover > hiscores_id;
                 hiscores_mover --) {
                this.hiscores[hiscores_mover] = this.hiscores[hiscores_mover - 1];
            }
            this.hiscores[hiscores_id] = { score: this.cash, name: "" };
            score_added = true;
            break;
        }
    }
    if (!score_added && this.hiscores.length < this.HISCORES_COUNT) {
        this.hiscores.push({ score: this.cash, name: "" });
    } else if (this.hiscores.length > this.HISCORES_COUNT) {
        this.hiscores.length = this.HISCORES_COUNT;
    }

    this.cash = 0;
    this.credits = 0;

    this.constructHiscoreWidgets();

}

QtPayazzo.prototype.hiscoreGiven = function()
{
    if (this.hiscore_input.text == "") {
        return;
    }
    for (var hiscores_id = 0; hiscores_id < this.hiscores.length; hiscores_id ++) {
        var hiscore_row = this.hiscores[hiscores_id];
        if (hiscore_row.name == "") {
            hiscore_row.name = this.hiscore_input.text;
        }
    }
    this.constructHiscoreWidgets();
}

QtPayazzo.prototype.closeHiscore = function()
{
    this.mainlayout.removeWidget(this.layoutholder);
    this.layoutholder.setParent(null);
    this.layoutholder = null;
    this.hiscore_input = null;
}

QtPayazzo.prototype.constructHiscoreWidgets = function()
{
    if (this.layoutholder) {
        this.mainlayout.removeWidget(this.layoutholder);
        this.layoutholder.setParent(null);
    }
    var userlayout = new QVBoxLayout();
    userlayout.addStretch();
    userlayout.addWidget(new QLabel("High scores:", this), 0, Qt.AlignCenter);
    var hiscores_layout = new QHBoxLayout();
    var hiscore_layout1 = new QFormLayout();
    var hiscore_layout2 = new QFormLayout();
    this.hiscore_input = null;
    for (var hiscores_id = 0; hiscores_id < this.hiscores.length; hiscores_id ++) {
        var hiscore_row = this.hiscores[hiscores_id];
        if (hiscore_row.name == "") {
            this.hiscore_input = new QLineEdit(this);
            this.hiscore_input.maxLength = 12;
            this.hiscore_input["returnPressed"].connect(this, this.hiscoreGiven);
            if (hiscores_id < this.hiscores.length / 2) {
                hiscore_layout1.addRow(hiscore_row.score.toString(), this.hiscore_input);
            } else {
                hiscore_layout2.addRow(hiscore_row.score.toString(), this.hiscore_input);
            }
        } else {
            if (hiscores_id < this.hiscores.length / 2) {
                hiscore_layout1.addRow(hiscore_row.score.toString(), new QLabel(hiscore_row.name, this));
            } else {
                hiscore_layout2.addRow(hiscore_row.score.toString(), new QLabel(hiscore_row.name, this));
            }
        }
    }
    hiscores_layout.addStretch();
    hiscores_layout.addLayout(hiscore_layout1);
    hiscores_layout.addStretch();
    hiscores_layout.addLayout(hiscore_layout2);
    hiscores_layout.addStretch();
    userlayout.addLayout(hiscores_layout)
    var closebutton = new QPushButton("Close", this);
    if (this.hiscore_input) {
        closebutton.setEnabled(false);
    }
    closebutton["clicked"].connect(this, this.closeHiscore);
    userlayout.addWidget(closebutton, 0, Qt.AlignCenter);
    userlayout.addStretch();
    this.layoutholder = new LayoutHolder(this, userlayout);
    this.mainlayout.addWidget(this.layoutholder, 0, 0);
}

QtPayazzo.prototype.fileLoaded = function(caller, filename, data, filegroup)
{
    var new_pixmap = new QPixmap();
    new_pixmap.loadFromData(data, 0, Qt.AutoColor);
    if (filename == "http://lively.cs.tut.fi/images/payazzo/Payazzo.png") {
        caller.pixmap_background = new_pixmap;
    } else if (filename == "http://lively.cs.tut.fi/images/payazzo/PayazzoMask.png") {
        caller.pixmap_mask = new_pixmap;
    } else if (filename == "http://lively.cs.tut.fi/images/payazzo/Coin.png") {
        caller.pixmap_coin = new_pixmap;
    } else if (filename == "http://lively.cs.tut.fi/images/payazzo/5Coins.png") {
        caller.pixmap_5coins = new_pixmap;
    } else if (filename == "http://lively.cs.tut.fi/images/payazzo/16Coins.png") {
        caller.pixmap_16coins = new_pixmap;
    } else if (filename == "http://lively.cs.tut.fi/images/payazzo/NewGame.png") {
        caller.pixmap_newgame = new_pixmap;
    } else if (filename == "http://lively.cs.tut.fi/images/payazzo/InsertCoinHover.png") {
        caller.pixmap_insertcoin_hover = new_pixmap;
    } else if (filename == "http://lively.cs.tut.fi/images/payazzo/InsertCoinPress.png") {
        caller.pixmap_insertcoin_press = new_pixmap;
    } else if (filename == "http://lively.cs.tut.fi/images/payazzo/CashOutHover.png") {
        caller.pixmap_cashout_hover = new_pixmap;
    } else if (filename == "http://lively.cs.tut.fi/images/payazzo/CashOutPress.png") {
        caller.pixmap_cashout_press = new_pixmap;
    }
    // Check if this was last file to get
    if (filegroup.files_left == 0) {
        caller.fileloader = null;
        caller.files_loaded = true;
        caller.update();
    }
}

QtPayazzo.prototype.paintEvent = function(event)
{

    // If files are not loaded, then do not paint anything.
    if (!this.files_loaded) {
        return;
    }

    var painter = new QPainter();

    painter.begin(this);
    painter.setRenderHint(QPainter.Antialiasing);
    painter.setClipRect(0, 0, 320, 480);

    // Draw background
    painter.drawPixmap(new QPoint(0, 0), this.pixmap_background);

    // Draw scores
    painter.setPen(new QColor(255, 0, 0));
    painter.drawText(new QPoint(103, 29), "Credits: " + this.credits);
    painter.drawText(new QPoint(15, 29), "Cash: " + this.cash);

    // Draw special buttons
    if (this.cashout == 1) {
        painter.drawPixmap(new QPoint(198, 12), this.pixmap_cashout_hover);
    } else if (this.cashout == 3) {
        painter.drawPixmap(new QPoint(198, 12), this.pixmap_cashout_press);
    }
    if (this.insertcoin == 1) {
        painter.drawPixmap(new QPoint(284, 12), this.pixmap_insertcoin_hover);
    } else if (this.insertcoin == 3) {
        painter.drawPixmap(new QPoint(284, 12), this.pixmap_insertcoin_press);
    }

    // Draw coins
    for (var coins_id = 0;
         coins_id < this.coins.length;
         coins_id ++) {
        coin = this.coins[coins_id];
        // The idea of the following is to reduce the shaking of circle.
        if (Math.abs(coin.render_x - coin.x) > 0.5 ||
            Math.abs(coin.render_y - coin.y) > 0.5) {
            coin.render_x = coin.x;
            coin.render_y = coin.y;
        }
        painter.drawPixmap(new QPoint(coin.render_x - 5, coin.render_y - 5), this.pixmap_coin);
    }

    // Draw coin stacks
    for (var stoppers_id = 0; stoppers_id < 12; stoppers_id ++) {
        var x = (stoppers_id - 5.5) * 14 + 160 - 5;
        var y = 246 + Math.abs(stoppers_id - 5.5) * 10 - 5;
        painter.drawPixmap(new QPoint(x, y + this.stoppers[stoppers_id].getPixelSpace()), this.pixmap_16coins);
    }
    painter.drawPixmap(new QPoint(13-5, 320-10*5-5 + this.stoppers[12].getPixelSpace()), this.pixmap_5coins);
    painter.drawPixmap(new QPoint(320-13-5, 320-10*5-5 + this.stoppers[13].getPixelSpace()), this.pixmap_5coins);

    // Draw mask
    painter.drawPixmap(new QPoint(0, 300), this.pixmap_mask);

    // If there is layout over the play area, draw transparent background for
    // it.
    if (this.layoutholder) {
        var brush = new QBrush(new QColor(239, 239, 239, 224));
        painter.fillRect(0, 0, 320, 480, brush);
    }
    // If game is not started, then display new game image
    else if (!this.game_running) {
        painter.drawPixmap(new QPoint(0, 0), this.pixmap_newgame);
    }

    painter.end();
}

/* To use the application in standalone mode, remove the comment around this code
var payazzo = new QtPayazzo();

payazzo.show();
QCoreApplication.exec();
*/
