/*========================================================================
 * Copyright (c) 2009 Tampere University of Technology, Finland
 * 
 * Permission is hereby granted, free of charge, to any person 
 * obtaining a copy of this software and associated documentation 
 * files (the "Software"), to deal in the Software without restriction, 
 * including without limitation the rights to use, copy, modify, merge, 
 * publish, distribute, sublicense, and/or sell copies of the Software, 
 * and to permit persons to whom the Software is furnished to do so, 
 * subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.  
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY 
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *======================================================================*/


// QtPlayer is a simple music player utilizing the Phonon multimedia 
// framework. In order to be used install the required Phonon libraries 
// to your system (see the QtSoftware website for reference). 
// Then add qt.phonon to the extensions, when compiling qtbindings and the 
// qs_eval binary.
// Besides being a conventional music player, the application 
// shows an image of the artist if available. Images are fetched from 
// last.fm music service. 
// 
// At the moment only the basic functionality is included in the player. 
// The player could be easily extended to support more advanced functions such 
// as show more detailed information about the artist, show current gigs on the 
// map etc. The support for playing lists or queues should be added as well. 
// When used within the Lively for Qt, and not in the stand-alone mode, there 
// are problems with Mac OS X as the qs_eval menu overrides the QtPlayer manu. 
// This should be fixed (the UI could be redesigned).

function tr(s) {
    return s;
}


function containsPhononSupport() {
	try {
		// Try to create a player instance
		Phonon.createPlayer(Phonon.MusicCategory, null);	
	} 
	// Show a message for the user, if the support for Phonon is missing. 
	catch(err) {
    	var msg = new QMessageBox();
		msg.text='Phonon is not supported by your system. Make sure your system contains Phonon libraries and that the Phonon support has been compiled to qs_eval as well.';
    	msg.show();	
		return false;
	}
	return true;
}


function QtPlayer() {
    QMainWindow.call(this);

	// Try to create a player instance
	this.musicPlayer = Phonon.createPlayer(Phonon.MusicCategory, this.playTitle);	

 	// used for networking functions
	this.accessManager = new QNetworkAccessManager();
	// Setup the needed actions, menus and UI
	this.setupActions();
	this.setupMenus();
	this.setupUi();

	// The LCD is updated once per second.
	this.musicPlayer.setTickInterval(1000);
	// Update the artist and title info, when the metadata changes (title changes).
	this.musicPlayer['metaDataChanged()'].connect(this, this.updateTitleInfo);
	// Update the LCD display
	this.musicPlayer.tick.connect(this, this.tick); 
}




QtPlayer.prototype = new QMainWindow();




QtPlayer.prototype.about = function() {
    QMessageBox.information(this, tr('About Music Player'),
        tr('A simple Music Player implemented by utilizing QtScript and Phonon Multimedia Framework. The player contains only simple playback operations. If title metadata is available, the application attempts to download and show an image of the artist and the lyrics of the current track.'));
};




QtPlayer.prototype.openFile = function() {
	// Open a file
    var fileName = QFileDialog.getOpenFileNames(this, tr('Select File'),
        QDesktopServices.storageLocation(QDesktopServices.MusicLocation));
    // Create a new MediaSource of the chosen file
	this.src = new MediaSource(fileName.toString());
	// Set file as source
	this.musicPlayer.setCurrentSource(this.src);
	this.currTimeLcd.display('00:00');
	// Enable the play button after the file has been opened
    this.playAction.setDisabled(false);
};




// Actions for playback, pause, stop and exit
QtPlayer.prototype.playIt = function() {
    this.musicPlayer.play();
	this.playAction.setEnabled(false);
	this.pauseAction.setEnabled(true);
	this.stopAction.setEnabled(true);
};




QtPlayer.prototype.pauseIt = function() {
    this.musicPlayer.pause();
	this.playAction.setEnabled(true);
	this.pauseAction.setEnabled(false);
	this.stopAction.setEnabled(true);
};




QtPlayer.prototype.stopIt = function() {
	this.musicPlayer.stop();
	this.playAction.setEnabled(true);
	this.pauseAction.setEnabled(false);
	this.stopAction.setEnabled(false);        
	this.currTimeLcd.display('00:00');
};




QtPlayer.prototype.exitApp = function() {
	QCoreApplication.exit();
};




// Updates the time displayed on the LCD
QtPlayer.prototype.tick = function() {
	var currTime = new QTime(0, (this.musicPlayer.currentTime() / 60000) % 60, (this.musicPlayer.currentTime() / 1000) % 60);
	// Convert time to string and to be displayed on the LCD
	this.currTimeLcd.display(currTime.toString('mm:ss'));
};




// Handles updating the display with the new metadata, when the title is changed.
QtPlayer.prototype.updateTitleInfo = function() {

	// Fetch the name of the artist, title and album from the metadata 
	this.artist = this.musicPlayer.metaData(Phonon.ArtistMetaData);
	this.title = this.musicPlayer.metaData(Phonon.TitleMetaData);
	this.album = this.musicPlayer.metaData(Phonon.AlbumMetaData);
	
	if ( this.artist.length === 0 ) {
		this.artist = 'Unknown Artist';
	}
	if ( this.title.length === 0 ) {
		this.title = 'Unknown Title';
	}
	if ( this.album.length === 0 ) {
		this.album = 'Unknown Album';
	}

	// Get the total length of the track
	var total = this.musicPlayer.totalTime();
	// QTime ( int h, int m, int s = 0, int ms = 0 )
	this.titleLength = new QTime(0, (total / 60000) % 60, (total / 1000) % 60);
	
	// Set the artist, title and the total time of the track
	this.musicTitleLbl.text = '<h3>' +this.title +'</h3><h4>' +this.artist +'</h4>(on <i>' + this.album +'</i>)';
	
	// a new QUrl is formed for fetching an artist image
	var artistImageUrl = new QUrl(this.getArtistImageUrl(this.artist));
	// The reply contains the fetched page with an artist image
	this.imageReply = this.downloadFile(artistImageUrl, this.imageDownloaded);
};




// Connect buttons and menus with actions 
QtPlayer.prototype.setupActions = function() {

    this.playAction = new QAction(this.style().standardIcon(QStyle.SP_MediaPlay), tr('Play'), this);
    this.playAction.setDisabled(true);
    this.playAction.triggered.connect(this, this.playIt);

    this.pauseAction = new QAction(this.style().standardIcon(QStyle.SP_MediaPause), tr('Pause'), this);
    this.pauseAction.setDisabled(true);
    this.pauseAction.triggered.connect(this, this.pauseIt);

    this.stopAction = new QAction(this.style().standardIcon(QStyle.SP_MediaStop), tr('Stop'), this);
    this.stopAction.setDisabled(true);
    this.stopAction.triggered.connect(this, this.stopIt);

    this.openFileAction = new QAction(tr('&Open file'), this);
    this.openFileAction.triggered.connect(this, this.openFile);

    this.aboutAction = new QAction(tr('About'), this);
    this.aboutAction.triggered.connect(this, this.about);

    this.aboutQtAct = new QAction(tr('About &Qt'), this);
    this.aboutQtAct.triggered.connect(qApp.aboutQt);

    this.exitAction = new QAction(tr('E&xit'), this);
    this.exitAction.triggered.connect(this, this.exitApp);
};




// Add the menu items and add actions to them
QtPlayer.prototype.setupMenus = function() {
    this.fileMenu = this.menuBar().addMenu(tr('&File'));

    // ### working around bug in QMenu.prototype.addAction
    QMenu.prototype.addAction = QWidget.prototype.addAction;

    this.fileMenu.addAction(this.openFileAction);
    this.fileMenu.addSeparator();
    this.fileMenu.addAction(this.exitAction);

    this.aboutMenu = this.menuBar().addMenu(tr('&Help'));
    this.aboutMenu.addAction(this.aboutAction);
    this.aboutMenu.addAction(this.aboutQtAct);
};




// Sets up the toolbars, layouts etc.
QtPlayer.prototype.setupUi = function() {

    this.bar = new QToolBar(this);
    QToolBar.prototype.addAction = QWidget.prototype.addAction;
    this.bar.addAction(this.playAction);
    this.bar.addAction(this.pauseAction);
    this.bar.addAction(this.stopAction);
	this.styleSheet = 'QMainWindow{ background-color: white};';

	// For showing information of the name of the artist and the title to be played 
	this.musicTitleLbl = new QLabel();
	// The artist image is loaded to a QPixMap
	this.artistImage = new QPixmap();
	// The image is placed to a QLabel
	this.artistImageLbl = new QLabel();

	this.seekSlider = new SeekSlider(this.musicPlayer, this);
	this.seekSlider.iconVisible = true;
    var palette = new QPalette();
    palette.setBrush(QPalette.Light, Qt.darkGrey);
    this.currTimeLcd = new QLCDNumber();
    this.currTimeLcd.display('00:00');
	this.currTimeLcd.palette = palette;
	this.currTimeLcd.lineWidth = 0;
	
	// the current title and artist image
	this.titleLayout = new QHBoxLayout();
	this.titleLayout.addWidget(this.musicTitleLbl,0,0);
	this.titleLayout.addWidget(this.artistImageLbl,0,0);
	
	// Add the title data and lyrics to the layout containing title data
	this.dataLayout = new QVBoxLayout();
	this.dataLayout.addLayout(this.titleLayout,0);

	// Playbacklayout contains the playback bar with play, pause and stop buttons
    this.playbackLayout = new QHBoxLayout();
    this.playbackLayout.addWidget(this.bar,0,0);

	// Sliderlayout contains a seekslider plus an LCD display
	this.sliderLayout = new QHBoxLayout();
	this.sliderLayout.addWidget(this.seekSlider,0,0);
	this.sliderLayout.addWidget(this.currTimeLcd,0,0);

	// Add the other layouts to the mainlayout
    this.mainLayout = new QVBoxLayout();
    this.mainLayout.addLayout(this.dataLayout); // track info etc.
	this.mainLayout.addLayout(this.sliderLayout); // slider and LCD
    this.mainLayout.addLayout(this.playbackLayout); // play, stop and pause buttons

    this.widget = new QWidget();
    this.widget.setLayout(this.mainLayout);

    this.setCentralWidget(this.widget);
    this.resize(400,250);
    this.setWindowTitle('QtPlayer');
};




// Forms a URL to be used at Last.fm to fetch artist images.
// The request is of format http://ws.audioscrobbler.com/2.0/?method=artist.getimages&artist=cher&api_key=b25b959554ed76058ac220b7b2e0a026
QtPlayer.prototype.getArtistImageUrl = function(artist) {
	var baseUrl = 'http://ws.audioscrobbler.com/2.0/?method=artist.getimages&artist=';
	// Get your own API key at Last.fm and replave the one below with that!
	var APIKey = 'dc369d73f68488889ef64ba06686f6dd'; 
	var imgUrl = baseUrl + artist + '&api_key=' +APIKey;
	return imgUrl;
};





// Downloads the file pointed by the url
QtPlayer.prototype.downloadFile = function(url, callback) {

    // a QNetworkAccessManager handles sending requests and getting replies
    var accessManager = new QNetworkAccessManager();
    // An HTTP request is sent 
    var reqReply = accessManager.get(new QNetworkRequest(url));
    // When the reply is obtained, the execution proceeds to the callback function.
    reqReply.finished.connect(this, callback);
    // the reply is returned
    return reqReply;
};




// Process the downloaded artist image (obtained from Last.fm).
// The link is contained within an XML document.
QtPlayer.prototype.imageDownloaded = function() {

    // Create a data structure to store the read bytes
    var reply = new QByteArray();
    // Read the data from the reply to a QByteArray
    reply = this.imageReply.readAll();
    // New QXmlStreamReader for reading the XML file.
    var xml = new QXmlStreamReader(reply);

	// Create an array for image links
	var imgLinks = [];
	
    // Read the file, until the end is reached.
    while (!xml.atEnd()) {
        xml.readNext();
        if (xml.isStartElement()) {
            if (xml.name() == 'size') {
    
				// Obtain the attrubutes of the element.
				var attrs = xml.attributes();
				var orig = attrs.at(0);
				
				// Use only images of original-size. 
				// This could be optimized as there are smaller images available.
				if ( orig.value() == 'original' ) {
		            var imgLink = xml.readElementText();
		            imgLinks.push(imgLink);
				}
            }
        }
   }


	// Sets the image of the artist to a QLabel
	function showImg() {
		var data = reqReply.readAll();
		this.artistImage.loadFromData(data, 0, Qt.AutoColor);
		this.artistImage = this.artistImage.scaled(150, 150, Qt.KeepAspectRatio, Qt.SmoothTransformation);
		// Set the downloaded image to the QLabel
		this.artistImageLbl.setPixmap(this.artistImage);
	}

	// If images were found, choose an image on basis of a draw. 
	if ( imgLinks.length !== 0 ) {

		var index = Math.floor(Math.random() * imgLinks.length);
		// Send an HTTP/GET request to fetch the image
		var reqReply = this.accessManager.get(new QNetworkRequest(new QUrl(imgLinks[index])));
		// When a reply has been received, proceed to show the image.
		reqReply.finished.connect(this, showImg);
	}
};




/* To use the application in standalone mode, remove the comment around this code 
if ( containsPhononSupport() === true ) {
	var wdgt = new QtPlayer();
	wdgt.show();
	QCoreApplication.exec();
}*/

