/*========================================================================
 * Copyright (c) 2009 Tampere University of Technology, Finland
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without restriction,
 * including without limitation the rights to use, copy, modify, merge,
 * publish, distribute, sublicense, and/or sell copies of the Software,
 * and to permit persons to whom the Software is furnished to do so,
 * subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *======================================================================*/

function PhysicsCollision(normal, depth)
{
    this.normal = normal;
    this.depth = depth;
    this.really_hits = false;
}

function PhysicsCircle(x, y, radius, xv, yv, visible)
{
    this.x = x;
    this.y = y;
    this.radius = radius;
    this.xv = xv;
    this.yv = yv;
    this.visible = visible;
}

function PhysicsPolygon(points)
{
    this.polygon = new QPolygonF(points);
    this.bounds = this.polygon.boundingRect();
}

function PhysicsPath(points, disable_begin, disable_end, closed)
{
    this.points = points;
    this.disable_begin = this.disable_begin;
    this.disable_end = this.disable_end;
    this.closed = closed;
    this.bounds = new QPolygonF(points).boundingRect();
}

function PhysicsSimulation()
{
    this.reset();
}

PhysicsSimulation.prototype.run = function(secm)
{
    var GRAVITY = 40;
    for (var actors_id = 0;
         actors_id < this.actors.length;
         actors_id ++) {
        var circle = this.actors[actors_id];

        // Check if actor moves so fast, that it is needed to be ran in parts
        var parts = Math.ceil(Math.sqrt(circle.xv*circle.xv + circle.yv*circle.yv)*secm / (circle.radius));
        if (parts == 0) {
            parts = 1;
        }
        var secm2 = secm / parts;

        for (var parts_id = 0;
             parts_id < parts;
             parts_id ++) {

            // Run basic physics
            circle.x += circle.xv * secm2;
            circle.y += circle.yv * secm2;
            circle.yv += GRAVITY * secm2;

            // Check collisions to other actors
            for (var actors_id2 = actors_id+1;
                 actors_id2 < this.actors.length;
                 actors_id2 ++) {
                var circle2 = this.actors[actors_id2];
                var diff_x = circle2.x - circle.x;
                var diff_y = circle2.y - circle.y;
                var diff = Math.sqrt(diff_x*diff_x + diff_y*diff_y);
                var both_radius = circle.radius + circle2.radius;
                if (diff < both_radius) {
                    var m = (both_radius - diff) / both_radius / 2.0;
                    circle.x -= diff_x*m;
                    circle.y -= diff_y*m;
                    circle2.x += diff_x*m;
                    circle2.y += diff_y*m;
                    var coll_normal_x = diff_x / diff;
                    var coll_normal_y = diff_y / diff;
                    this.modifyVelocityUsingCollision(circle, -coll_normal_x, -coll_normal_y);
                    this.modifyVelocityUsingCollision(circle2, coll_normal_x, coll_normal_y);
                }
            }

            var EXTRA_DETECTION_RANGE = 2.5;

            // Some helper variables
            var circle_radius_pow_2 = (circle.radius+EXTRA_DETECTION_RANGE)*(circle.radius+EXTRA_DETECTION_RANGE);

            // Check collisions to obstacles
            var colls = [];
            var colls_deepest_id;
            var colls_deepest_depth = -9999;
            // Polygons
            for (var polygons_id = 0;
                 polygons_id < this.polygons.length;
                 polygons_id ++) {
                var poly_cobj = this.polygons[polygons_id];
                var polygon = poly_cobj.polygon;
                // Do bounding box test and ignore polygons that has their
                // bounding box out of actors one.
                if (poly_cobj.bounds.left() > circle.x + circle.radius + EXTRA_DETECTION_RANGE ||
                    poly_cobj.bounds.right() < circle.x - circle.radius - EXTRA_DETECTION_RANGE ||
                    poly_cobj.bounds.top() > circle.y + circle.radius + EXTRA_DETECTION_RANGE ||
                    poly_cobj.bounds.bottom() < circle.y - circle.radius - EXTRA_DETECTION_RANGE) {
                    continue;
                }
                for (var edge_id = 0;
                     edge_id < polygon.size();
                     edge_id ++) {
                    var begin = polygon.at(edge_id);
                    var end = polygon.at((edge_id + 1) % polygon.size());
                    var edge_x = end.x() - begin.x();
                    var edge_y = end.y() - begin.y();

                    // Calculate how much the circle should be moved so that line
                    // segment from it to the beginning of polygon line would be 90
                    // degrees from polygonline. Note that the unit here is -edge,
                    // so this whole thing will tell how much the moving must be
                    // done using negative variable "edge" as a mover vector.
                    var m = ((circle.x - begin.x())*edge_x + (circle.y - begin.y())*edge_y) /
                            (edge_x*edge_x + edge_y*edge_y);
                    // Check if circle touches this edge.
                    if (m >= 0.0 && m <= 1.0) {
                        var diff_x = circle.x - edge_x * m - begin.x();
                        var diff_y = circle.y - edge_y * m - begin.y();
                        var diff_pow_2 = diff_x*diff_x + diff_y*diff_y;
                        if (diff_pow_2 <= circle_radius_pow_2) {
                            // Calculate distance and then form normal at collision
                            // point.
                            var diff = Math.sqrt(diff_pow_2);
                            if (diff != 0) {
                                var depth = circle.radius - diff;
                                colls.push(new PhysicsCollision(new QPointF(diff_x / diff, diff_y / diff), depth));
                                if (depth > colls_deepest_depth) {
                                    colls_deepest_depth = depth;
                                    colls_deepest_id = colls.length - 1;
                                }
                            }
                        } else {
                            diff_x = circle.x - begin.x();
                            diff_y = circle.y - begin.y();
                            diff_pow_2 = diff_x*diff_x + diff_y*diff_y;
                            if (diff_pow_2 <= circle_radius_pow_2) {
                                // Calculate distance and then form normal at collision
                                // point.
                                var diff = Math.sqrt(diff_pow_2);
                                if (diff != 0) {
                                    var depth = circle.radius - diff;
                                    colls.push(new PhysicsCollision(new QPointF(diff_x / diff, diff_y / diff), depth));
                                    if (depth > colls_deepest_depth) {
                                        colls_deepest_depth = depth;
                                        colls_deepest_id = colls.length - 1;
                                    }
                                }
                            }
                        }
                    } else {
                        var diff_x = circle.x - begin.x();
                        var diff_y = circle.y - begin.y();
                        var diff_pow_2 = diff_x*diff_x + diff_y*diff_y;
                        if (diff_pow_2 <= circle_radius_pow_2) {
                            // Calculate distance and then form normal at collision
                            // point.
                            var diff = Math.sqrt(diff_pow_2);
                            if (diff != 0) {
                                var depth = circle.radius - diff;
                                colls.push(new PhysicsCollision(new QPointF(diff_x / diff, diff_y / diff), depth));
                                if (depth > colls_deepest_depth) {
                                    colls_deepest_depth = depth;
                                    colls_deepest_id = colls.length - 1;
                                }
                            }
                        }
                    }

                }
            }
            // Paths
            for (var paths_id = 0;
                 paths_id < this.paths.length;
                 paths_id ++) {
                var path_cobj = this.paths[paths_id];
                var path = path_cobj.points;
                // Do bounding box test and ignore paths that has their
                // bounding box out of actors one.
                if (path_cobj.bounds.left() > circle.x + circle.radius + EXTRA_DETECTION_RANGE ||
                    path_cobj.bounds.right() < circle.x - circle.radius - EXTRA_DETECTION_RANGE ||
                    path_cobj.bounds.top() > circle.y + circle.radius + EXTRA_DETECTION_RANGE ||
                    path_cobj.bounds.bottom() < circle.y - circle.radius - EXTRA_DETECTION_RANGE) {
                    continue;
                }
                var end_edge_id;
                if (path_cobj.closed) {
                    end_edge_id = path.length;
                } else {
                    end_edge_id = path.length - 1;
                }
                for (var edge_id = 0;
                     edge_id < end_edge_id;
                     edge_id ++) {
                    var begin = path[edge_id];
                    var end = path[(edge_id + 1) % path.length];
                    var edge_x = end.x() - begin.x();
                    var edge_y = end.y() - begin.y();

                    // Calculate how much the circle should be moved so that line
                    // segment from it to the beginning of path line would be 90
                    // degrees from pathline. Note that the unit here is -edge,
                    // so this whole thing will tell how much the moving must be
                    // done using negative variable "edge" as a mover vector.
                    var m = ((circle.x - begin.x())*edge_x + (circle.y - begin.y())*edge_y) /
                            (edge_x*edge_x + edge_y*edge_y);
                    // Check if circle touches this edge.
                    if (m >= 0.0 && m <= 1.0) {
                        var diff_x = circle.x - edge_x * m - begin.x();
                        var diff_y = circle.y - edge_y * m - begin.y();
                        var diff_pow_2 = diff_x*diff_x + diff_y*diff_y;
                        if (diff_pow_2 <= circle_radius_pow_2) {
                            // Calculate distance and then form normal at collision
                            // point.
                            var diff = Math.sqrt(diff_pow_2);
                            if (diff != 0) {
                                var depth = circle.radius - diff;
                                colls.push(new PhysicsCollision(new QPointF(diff_x / diff, diff_y / diff), depth));
                                if (depth > colls_deepest_depth) {
                                    colls_deepest_depth = depth;
                                    colls_deepest_id = colls.length - 1;
                                }
                            }
                        } else {
                            diff_x = circle.x - begin.x();
                            diff_y = circle.y - begin.y();
                            diff_pow_2 = diff_x*diff_x + diff_y*diff_y;
                            if (diff_pow_2 <= circle_radius_pow_2) {
                                // Calculate distance and then form normal at collision
                                // point.
                                var diff = Math.sqrt(diff_pow_2);
                                if (diff != 0) {
                                    var depth = circle.radius - diff;
                                    colls.push(new PhysicsCollision(new QPointF(diff_x / diff, diff_y / diff), depth));
                                    if (depth > colls_deepest_depth) {
                                        colls_deepest_depth = depth;
                                        colls_deepest_id = colls.length - 1;
                                    }
                                }
                            }
                        }
                    } else {
                        if (edge_id < end_edge_id - 1 ||
                            poly_cobj.closed ||
                           !poly_cobj.disable_end) {
                            var diff_x = circle.x - end.x();
                            var diff_y = circle.y - end.y();
                            var diff_pow_2 = diff_x*diff_x + diff_y*diff_y;
                            if (diff_pow_2 <= circle_radius_pow_2) {
                                // Calculate distance and then form normal at collision
                                // point.
                                var diff = Math.sqrt(diff_pow_2);
                                if (diff != 0) {
                                    var depth = circle.radius - diff;
                                    colls.push(new PhysicsCollision(new QPointF(diff_x / diff, diff_y / diff), depth));
                                    if (depth > colls_deepest_depth) {
                                        colls_deepest_depth = depth;
                                        colls_deepest_id = colls.length - 1;
                                    }
                                }
                            }
                        }
                        if (edge_id == 0 && !poly_cobj.closed) {
                            var diff_x = circle.x - begin.x();
                            var diff_y = circle.y - begin.y();
                            var diff_pow_2 = diff_x*diff_x + diff_y*diff_y;
                            if (diff_pow_2 <= circle_radius_pow_2) {
                                // Calculate distance and then form normal at collision
                                // point.
                                var diff = Math.sqrt(diff_pow_2);
                                if (diff != 0) {
                                    var depth = circle.radius - diff;
                                    colls.push(new PhysicsCollision(new QPointF(diff_x / diff, diff_y / diff), depth));
                                    if (depth > colls_deepest_depth) {
                                        colls_deepest_depth = depth;
                                        colls_deepest_id = colls.length - 1;
                                    }
                                }
                            }
                        }
                    }
                }
            }
            // Circles
            for (var circles_id = 0;
                 circles_id < this.circles.length;
                 circles_id ++) {
// TODO: Would bounding box here make things faster?
                var ob_circle = this.circles[circles_id];
                var diff_x = circle.x - ob_circle.x;
                var diff_y = circle.y - ob_circle.y;
                var diff = Math.sqrt(diff_x*diff_x + diff_y*diff_y);
                if (diff <= circle.radius + ob_circle.radius + EXTRA_DETECTION_RANGE) {
                    // Form normal at collision point.
                    if (diff != 0) {
                        var depth = circle.radius + ob_circle.radius - diff;
                        colls.push(new PhysicsCollision(new QPointF(diff_x / diff, diff_y / diff), depth));
                        if (depth > colls_deepest_depth) {
                            colls_deepest_depth = depth;
                            colls_deepest_id = colls.length - 1;
                        }
                    }
                }
            }
            if (colls.length > 0 && colls_deepest_depth > 0) {
                // Move actor out of obstacles. First actor is moved outside of
                // deepest collision and then actor is moved at the surface of
                // deepest collision so that it comes out of all other obstacles.
                var deepest_coll = colls[colls_deepest_id];
                var deepest_coll_xmod = deepest_coll.normal.x() * deepest_coll.depth;
                var deepest_coll_ymod = deepest_coll.normal.y() * deepest_coll.depth;
                circle.x += deepest_coll_xmod;
                circle.y += deepest_coll_ymod;
                deepest_coll.really_hits = true;

                // Go through other collisions and move actor out of them too.
                var colls_really_hit = 1;
                if (colls.length > 1) {
                    var move_x = deepest_coll.normal.y();
                    var move_y = -deepest_coll.normal.x();

                    var max_pos = 0;
                    var max_neg = 0;
                    for (var colls_id = 0; colls_id < colls.length; colls_id ++) {
                        // Skip the deepest collision.
                        if (colls_id == colls_deepest_id) {
                            continue;
                        }
                        var coll = colls[colls_id];
                        // The original collision has probably moved this
                        // collisions already, so it is good idea to calculate the
                        // effect of it.
                        var m = (coll.normal.x() * deepest_coll_xmod + coll.normal.y() * deepest_coll_ymod) /
                                (coll.normal.x() * coll.normal.x() + coll.normal.y() * coll.normal.y());
                        var depth = coll.depth - m;
                        if (depth > 0) {
                            colls_really_hit ++;
                            coll.really_hits = true;
                            // Calculate how much object needs to be moved using vector
                            // move so that it becomes out of the current obstacle.
                            var nrm_x = coll.normal.x() * depth;
                            var nrm_y = coll.normal.x() * depth;
                            m = (nrm_x*nrm_x + nrm_y*nrm_y) /
                                (nrm_x*move_x + nrm_y*move_y);
                            if (m > 0) {
                                if (max_pos < m) {
                                    max_pos = m;
                                }
                            } else {
                                if (max_neg > m) {
                                    max_neg = m;
                                }
                            }
                        }
                    }
                    if (max_pos > -max_neg) {
                        circle.x += move_x * max_pos;
                        circle.y += move_y * max_pos;
                    } else {
                        circle.x += move_x * max_neg;
                        circle.y += move_y * max_neg;
                    }
                }

                // Finally go through collisions as long as the velocity vector
                // points towards the collision surfaces.
                var colls_id = 0;
                var no_colls_found = 0;
                var tries = 100;
                do {
                    var coll = colls[colls_id];
                    if (coll.really_hits) {
                        if ((circle.xv*coll.normal.x() + circle.yv*coll.normal.y()) < 0) {
                            this.modifyVelocityUsingCollision(circle, coll.normal.x(), coll.normal.y());
                            no_colls_found = 0;
                        } else {
                            no_colls_found ++;
                        }
                        tries --;
                    }
                    colls_id ++;
                    if (colls_id == colls.length) {
                        colls_id = 0;
                    }
                } while (no_colls_found < colls_really_hit && tries > 0);

            }

        }

    }

}

PhysicsSimulation.prototype.modifyVelocityUsingCollision = function(actor, normal_x, normal_y)
{
    var BOUNCE_MULTIPLIER = 1.35;
    var dp = (normal_x*actor.xv + normal_y*actor.yv);
// TODO: Ensure this is a good idea to do!
    if (dp > 0) {
        actor.xv *= BOUNCE_MULTIPLIER;
        actor.yv *= BOUNCE_MULTIPLIER;
        return;
    }
    var m = -dp / (normal_x * normal_x + normal_y * normal_y);
    actor.xv += normal_x*m*BOUNCE_MULTIPLIER;
    actor.yv += normal_y*m*BOUNCE_MULTIPLIER;
}

PhysicsSimulation.prototype.reset = function()
{
    this.actors = [];
    this.polygons = [];
    this.paths = [];
    this.circles = [];
}

PhysicsSimulation.prototype.addCircle = function(cObj, active)
{
    if (active) {
        this.actors.push(cObj);
    } else {
        this.circles.push(cObj);
    }
}

PhysicsSimulation.prototype.addPolygon = function(polygon)
{
    this.polygons.push(polygon);
}

PhysicsSimulation.prototype.addPath = function(path)
{
    this.paths.push(path);
}

PhysicsSimulation.prototype.removeCircle = function(cObj, active)
{
    if (active) {
        for (var actors_id = 0; actors_id < this.actors.length; actors_id ++) {
            if (this.actors[actors_id] == cObj) {
                this.actors[actors_id] = this.actors[this.actors.length-1];
                this.actors.length --;
                return;
            }
        }
    } else {
        for (var circles_id = 0; circles_id < this.circles.length; circles_id ++) {
            if (this.circles[circles_id] == cObj) {
                this.circles[circles_id] = this.circles[this.circles.length-1];
                this.circles.length --;
                return;
            }
        }
    }
}
